# *-* coding=utf8 *-*
# ******************************************************************************
# helper.py
# ------------------------------------------------------------------------------
#
# Copyright (c) 2018 JSH <downloader@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
# ******************************************************************************
import pandas as pd
import matplotlib.pyplot as plt

from matplotlib.lines import Line2D
import matplotlib.ticker as tkr

from functools import reduce


"""Auflistung aller möglichen ECO Zustände mit HSI und Farbwert"""
#               TYPE       HSI    COLOR
_ECOTYPES = [("PWR 100%",   513, 'darkred'),
             ("PWR 75%",    400, 'red'),
             ("PWR 50%",    300, 'tomato'),
             ("PWR 25%",    200, 'salmon'),
             ("ECO 100%",   100, 'darkcyan'),
             ("ECO 75%",     75, 'lightseagreen'),
             ("ECO 50%",     50, 'cyan'),
             ("ECO 25%",     25, 'paleturquoise'),
             ("CHG LOW",      1, 'palegreen'),
             ("CHG MEDIUM",   0, 'green'),
             ("CHG HIGH",   -50, 'darkgreen'),
             ("BREAK",     -100, 'black')]


def _erstelle_verteilung(dataframe):
    """Ermittelt die Verteilung über alle ECO Typen

    :param dataframe:
    :return Dictionary mit der Verteilung
    """
    result = {}
    for ecotyp, _, _ in _ECOTYPES:
        result[ecotyp] = dataframe[dataframe.ECOTYP == ecotyp].COUNT.agg(sum)
    return result


class EcoTypVerteilung(object):
    """Erstellt Kreisdiagramme für die EcoTypes
    """

    def __init__(self, dataframe):
        """Initialisiert das Objekt

        :param dataframe: Enthält alle anzuzeigenen Werte
        """
        self._verteilung = _erstelle_verteilung(dataframe)
        self._gesamt = self._erzeuge_gesamt()
        self._chg = self._erzeuge_chg()
        self._eco = self._erzeuge_eco()
        self._pwr = self._erzeuge_pwr()
        self._chg_vs_break = self._erzeuge_chg_vs_break()

    def plot_gesamt(self):
        """Erzeugt ein Kreisdiagramm über den Gesamtbereich
        """
        colors = ('darkred', 'palegreen', 'cyan', 'red')
        self._plot(self._gesamt, colors)

    def plot_chg(self):
        """Erzeugt ein Kreisdiagramm über den CHG Bereich
        """
        colors = ('darkcyan', 'cyan', 'palegreen')
        self._plot(self._chg, colors)

    def plot_chg_vs_break(self):
        """Erzeugt ein Kreisdiagramm zum Verhältnis zwischen CHG und BREAK
        """
        colors = ('cyan', 'red')
        self._plot(self._chg_vs_break, colors)

    def plot_eco(self):
        """Erzeugt ein Kreisdiagramm über den ECO Bereich
        """
        colors = ('paleturquoise', 'cyan', 'lightseagreen', 'darkcyan')
        self._plot(self._erzeuge_eco(), colors)

    def plot_pwr(self):
        """Erzeugt ein Kreisdiagramm über den PWR Bereich
        """
        colors = ('tomato', 'red')
        self._plot(self._erzeuge_pwr(), colors)

    def _erzeuge_gesamt(self):
        """Ermittelt die Gesamtverteilung über alle EcoTypen

        :return Dictionary mit der Gesamtverteilung
        """
        v = self._verteilung
        result = {"BREAK": v["BREAK"],
                  "CHG": v["CHG HIGH"] + v["CHG MEDIUM"] + v["CHG LOW"],
                  "ECO": v["ECO 25%"] + v["ECO 50%"] + v["ECO 75%"] + v["ECO 100%"],
                  "PWR": v["PWR 25%"] + v["PWR 50%"] + v["PWR 75%"] + v["PWR 100%"]}
        return result

    def _erzeuge_chg(self):
        """Ermittelt die CHG Verteilung

        :return Dictionary mit der CHG Verteilung
        """
        v = self._verteilung
        return {
            "CHG HIGH":   v["CHG HIGH"],
            "CHG MEDIUM": v["CHG MEDIUM"],
            "CHG LOW":    v["CHG LOW"]
        }

    def _erzeuge_chg_vs_break(self):
        """Ermittelt das Verhältnis zwischen CHG und BREAK

        :return Dictionary mit dem CHG/BREAK Verhältnis
        """
        v = self._verteilung
        return {
            "CHG": v["CHG HIGH"] + v["CHG MEDIUM"] + v["CHG LOW"],
            "BREAK": v["BREAK"]
        }

    def _erzeuge_eco(self):
        """Ermittelt die ECO Verteilung

        :return Dictionary mit der ECO Verteilung
        """
        v = self._verteilung
        return {
            "bis ECO 25%": v["ECO 25%"],
            "bis ECO 50%": v["ECO 50%"],
            "bis ECO 75%": v["ECO 75%"],
            "bis ECO 100%": v["ECO 100%"],
        }

    def _erzeuge_pwr(self):
        """Ermittel die PWR Verteilung

        :return Dictionary mit der PWR Verteilung
        """
        v = self._verteilung
        return {
            "bis PWR 25%": v["PWR 25%"],
            "bis PWR 50%": v["PWR 50%"],
            "bis PWR 75%": v["PWR 75%"],
            "bis PWR 100%": v["PWR 100%"],
        }

    @staticmethod
    def _plot(data, colors):
        """Erstellt eine Kreisdiagramm mit den übergebenen Werten

        :param data: Dictionary: Alle anzuzeigenen Werte
        :param colors: Tuple: Zugeordnete Farben
        """
        _, ax1 = plt.subplots()
        data = _ignoriere_unter_ein_prozent(data)
        ax1.pie(data.values(), labels=data.keys(), colors=colors, autopct='%.01f%%')
        ax1.axis('equal')
        plt.show()


####
# Funktionen
####

def plot_fuelflowh_verteilung(dataframe, von=0, bis=20):
    """Erzeugt ein Histrogramm über die FUELFLOWH Werte
    
    :param dataframe: Enthält alle anzuzeigenen Werte
    :param von: Kleinster FUELFLOWH Wert
    :param bis: Größter FUELFLOWH Wert
    """
    title = "Durchschnittlicher Verbrauch"
    plt.hist(
        dataframe[(dataframe.FUELFLOWH >= von) & (dataframe.FUELFLOWH <= bis)].FUELFLOWH,
        bins=_bins(von, bis)
    )
    plt.title(title)
    plt.xlabel('FUELFLOWH')
    plt.ylabel('Anzahl Einträge')
    plt.grid(True)
    plt.show()


def plot_hsi_verbrauch(dataframe, von=-512, bis=512):
    """Erzeugt ein Scatter Diagramm über die HSI Werte und Verbrauch
    
    :param dataframe: Enhält alle anzuzeigenen Werte
    :param von: Kleinster HSI Wert (Default: -512)
    :param bis: Größter HSI Wert (Default: 512)
    """
    assert von <= bis, "Ungültiger Bereich!"

    fig, ax = plt.subplots()
    bereich = (dataframe.HSI >= von) & (dataframe.HSI <= bis)
    if 'COLOR' in dataframe:
        ax.scatter(
            x=dataframe[bereich].HSI,
            y=dataframe[bereich].FUELFLOWH,
            color=dataframe[bereich].COLOR
        )
    else:
        ax.scatter(x=dataframe[bereich].HSI, y=dataframe[bereich].FUELFLOWH)
    title = "Verbrauch im Bereich von {} bis {} HSI".format(von, bis)
    plt.title(title)
    plt.xlabel('HSI')
    plt.ylabel('Liter/100km')
    plt.grid(True)

    if 'COLOR' in dataframe:
        ecotypes = set(dataframe[bereich].ECOTYP)
        names = []
        lines = []
        for typ, _, color in _ECOTYPES:
            if typ in ecotypes:
                lines.append(Line2D([0], [0], color=color, lw=10))
                names.append(typ)
        ax.plot()
        ax.legend(lines, names)
    plt.show()


def plot_hsi_verteilung(dataframe, von=-512, bis=512):
    """Erzeugt ein Scatterdiagramm über alle HSI-Werte

    :param dataframe: Enhält alle anzuzeigenen Werte
    :param von: Kleinster HSI Wert (Default: -512)
    :param bis: Größter HSI Wert (Default: 512)
    """
    assert von < bis, "Ungültiger Bereich!"
    keys = []
    values = []
    for hsi in range(von, bis):
        value = list(dataframe[dataframe.HSI == hsi].groupby("HSI").size())
        if value:
            keys.append(hsi)
            values.append(value[0])
    _set_xy_formater()
    plt.scatter(x=keys, y=values)
    plt.title("HSI Verteilung")
    plt.xlabel('HSI')
    plt.ylabel('Anzahl Einträge')
    plt.grid(True)
    plt.show()


def berechne_mittelwert_und_reicher_daten_an(dataframe):
    """Berechnet Mittelwerte und ergänzt um EcoTyp, Count und Color
    
    :param dataframe: Enthält alle zu berechnenen Werte
    :return pd.DataFrame
    """
    verbrauch = {}
    for hsi, fuelflowh in dataframe.values:
        key = int(hsi)
        if key in verbrauch:
            verbrauch[key].append(fuelflowh)
        else:
            verbrauch[key] = [fuelflowh]
    keys = []; values = []; types = []; count = []; colors = []
    for key in verbrauch:
        keys.append(key)
        count.append(len(verbrauch[key]))
        values.append(reduce(lambda x, y: x + y, verbrauch[key]) / len(verbrauch[key]))
        types.append(_ecotyp(key))
        colors.append(_ecocolor(key))
    result = pd.DataFrame({
        'HSI': keys,
        'FUELFLOWH': values,
        'ECOTYP': types,
        'COUNT': count,
        'COLOR': colors})
    return result


####
# Helpers
####
def _ecotyp(hsi):
    """Liefert zum HSI Wert den zugehörigen EcoTyp
    
    :param hsi: HSI Wert
    :return EcoTyp als String
    """
    for typ, value, color in reversed(_ECOTYPES):
        if hsi < value:
            return typ


def _ecocolor(hsi):
    """Liefert zum HSI Wert den zugehörigen Farbwert
    
    :param hsi: HSI Wert
    :return Color
    """
    for typ, value, color in reversed(_ECOTYPES):
        if hsi < value:
            return color


def _set_xy_formater():
    """Setzt die Zahlendarstellung für die Achsen auf Integer
    """
    def func(x, _):
        return int(x)

    _, ax = plt.subplots()
    formater = tkr.FuncFormatter(func)
    ax.yaxis.set_major_formatter(formater)
    ax.xaxis.set_major_formatter(formater)


def _bins(von, bis):
    """Berechnet die Anzahl nötiger Zwischenschritte für Histogramme

    :param von: Beginn des Bereichs
    :param bis: Ende des Bereichs
    :return: Anzahl Zwischenschritte
    """
    assert von < bis, "Ungültige Parameter!"
    if von < 0:
        result = abs(von) + bis
    else:
        result = bis - von
    return result + 1


def _ignoriere_unter_ein_prozent(data):
    """Entfernt alle Einträge im Dictionary die unter 1% liegen
    
    :param data: Dictionary
    :return Dictionary
    """
    result = {}
    gesamt = 0
    for key in data:
        gesamt += data[key]
    for key in data:
        if (data[key] / 100.0) >= 1.0:
            result[key] = data[key]
    return result
